@@ -18,6 +18,9 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#define _GNU_SOURCE
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <unistd.h>
@@ -25,9 +28,11 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
-#include "check_error.h"
-#include "list.h"
+#include <signal.h>
+
 #include "check.h"
+#include "check_error.h"
+#include "check_list.h"
 #include "check_impl.h"
 #include "check_msg.h"
 #include "check_log.h"
@@ -54,20 +59,33 @@ static void srunner_iterate_suites (SRunner *sr,
 static void srunner_run_tcase (SRunner *sr, TCase *tc);
 static int srunner_run_unchecked_setup (SRunner *sr, TCase *tc);
 static void srunner_run_unchecked_teardown (SRunner *sr, TCase *tc);
-static void tcase_run_checked_setup (TCase *tc);
+static TestResult * tcase_run_checked_setup (SRunner *sr, TCase *tc);
 static void tcase_run_checked_teardown (TCase *tc);
 static void srunner_iterate_tcase_tfuns (SRunner *sr, TCase *tc);
 static void srunner_add_failure (SRunner *sr, TestResult *tf);
-static TestResult *tcase_run_tfun_fork (TCase *tc, TF *tf);
-static TestResult *tcase_run_tfun_nofork (TCase *tc, TF *tf);
-static TestResult *receive_result_info_fork (char *tcname, int status);
-static TestResult *receive_result_info_nofork (char *tcname);
-static void set_fork_info (TestResult *tr, int status);
+static TestResult *tcase_run_tfun_fork (SRunner *sr, TCase *tc, TF *tf);
+static TestResult *tcase_run_tfun_nofork (SRunner *sr, TCase *tc, TF *tf);
+static TestResult *receive_result_info_fork (const char *tcname,
+                                             const char *tname,
+					     int status, int expected_signal);
+static TestResult *receive_result_info_nofork (const char *tcname,
+                                               const char *tname);
+static void set_fork_info (TestResult *tr, int status, int expected_signal);
 static void set_nofork_info (TestResult *tr);
 static char *signal_msg (int sig);
+static char *signal_error_msg (int signal_received, int signal_expected);
 static char *pass_msg (void);
 static char *exit_msg (int exitstatus);
-static int waserror (int status);
+static int waserror (int status, int expected_signal);
+
+#define MSG_LEN 100
+
+static pid_t current_test;
+
+static void sig_handler(int sig_nr)
+{
+  kill(current_test, SIGKILL);
+}
 
 static void srunner_run_init (SRunner *sr, enum print_output print_mode)
 {
@@ -106,26 +124,36 @@ static void srunner_iterate_suites (SRunner *sr,
       tc = list_val (tcl);
       srunner_run_tcase (sr, tc);
     }
+    
+    log_suite_end (sr, s);
   }
 }
 
 void srunner_run_all (SRunner *sr, enum print_output print_mode)
 {
+  struct sigaction old_action;
+  struct sigaction new_action;
+  
   if (sr == NULL)
     return;
   if (print_mode < 0 || print_mode >= CK_LAST)
     eprintf("Bad print_mode argument to srunner_run_all: %d",
 	    __FILE__, __LINE__, print_mode);
-      
+  
+  memset(&new_action, 0, sizeof new_action);
+  new_action.sa_handler = sig_handler;
+  sigaction(SIGALRM, &new_action, &old_action);
   srunner_run_init (sr, print_mode);
   srunner_iterate_suites (sr, print_mode);
   srunner_run_end (sr, print_mode);
+  sigaction(SIGALRM, &old_action, NULL);
 }
 
 static void srunner_add_failure (SRunner *sr, TestResult *tr)
 {  
   list_add_end (sr->resultlst, tr);
-  if (tr->ctx == CK_CTX_TEST)
+  /* If the context is either of these, the test has run. */
+  if ((tr->ctx == CK_CTX_TEST) || (tr->ctx == CK_CTX_TEARDOWN))
     sr->stats->n_checked++;
   if (tr->rtype == CK_FAILURE)
     sr->stats->n_failed++;
@@ -146,10 +174,10 @@ static void srunner_iterate_tcase_tfuns (SRunner *sr, TCase *tc)
     tfun = list_val (tfl);
     switch (srunner_fork_status(sr)) {
     case CK_FORK:
-      tr = tcase_run_tfun_fork (tc, tfun);
+      tr = tcase_run_tfun_fork (sr, tc, tfun);
       break;
     case CK_NOFORK:
-      tr = tcase_run_tfun_nofork (tc, tfun);
+      tr = tcase_run_tfun_nofork (sr, tc, tfun);
       break;
     default:
       eprintf("Bad fork status in SRunner", __FILE__, __LINE__);
@@ -171,37 +199,60 @@ static int srunner_run_unchecked_setup (SRunner *sr, TCase *tc)
   l = tc->unch_sflst;
 
   for (list_front(l); !list_at_end(l); list_advance(l)) {
-    
-    f = list_val(l);
     send_ctx_info(get_send_key(),CK_CTX_SETUP);
+    f = list_val(l);
     f->fun();
 
-    tr = receive_result_info_nofork (tc->name);
+    tr = receive_result_info_nofork (tc->name, "unchecked_setup");
 
     if (tr->rtype != CK_PASS) {
       srunner_add_failure(sr, tr);
       rval = 0;
       break;
     }
+    free(tr->file);
+    free(tr->msg);
+    free(tr);
   } 
 
   set_fork_status(srunner_fork_status(sr));
   return rval;
 }
 
-static void tcase_run_checked_setup (TCase *tc)
+static TestResult * tcase_run_checked_setup (SRunner *sr, TCase *tc)
 {
+  TestResult *tr = NULL;
   List *l;
   Fixture *f;
-
+  enum fork_status fstat = srunner_fork_status(sr);
+  
   l = tc->ch_sflst;
+  if (fstat == CK_FORK) {
+    send_ctx_info(get_send_key(),CK_CTX_SETUP);
+  }
   
-  send_ctx_info(get_send_key(),CK_CTX_SETUP);
-
   for (list_front(l); !list_at_end(l); list_advance(l)) {
+    if (fstat == CK_NOFORK) {
+      send_ctx_info(get_send_key(),CK_CTX_SETUP);
+    }
     f = list_val(l);
     f->fun();
+
+    /* Stop the setup and return the failure if nofork mode. */
+    if (fstat == CK_NOFORK) {
+      tr = receive_result_info_nofork (tc->name, "checked_setup");
+      if (tr->rtype != CK_PASS) {
+        break;
+      }
+
+      free(tr->file);
+      free(tr->msg);
+      free(tr);
+      tr = NULL;
+    }
   }
+
+  return tr;
 }
 
 static void tcase_run_checked_teardown (TCase *tc)
@@ -238,29 +289,31 @@ static void srunner_run_unchecked_teardown (SRunner *sr, TCase *tc)
 
 static void srunner_run_tcase (SRunner *sr, TCase *tc)
 {
-  
-  if (srunner_run_unchecked_setup (sr,tc)) {  
-  
+  if (srunner_run_unchecked_setup(sr,tc)) {
     srunner_iterate_tcase_tfuns(sr,tc);
-  
-    srunner_run_unchecked_teardown (sr, tc);
+    srunner_run_unchecked_teardown(sr, tc);
   }
 }
 
-static TestResult *receive_result_info_fork (char *tcname, int status)
+static TestResult *receive_result_info_fork (const char *tcname,
+                                             const char *tname,
+					     int status, int expected_signal)
 {
   TestResult *tr;
 
-  tr = receive_test_result (get_recv_key(), waserror(status));
+  tr = receive_test_result(get_recv_key(),
+                           waserror(status, expected_signal));
   if (tr == NULL)
     eprintf("Failed to receive test result", __FILE__, __LINE__);
   tr->tcname = tcname;
-  set_fork_info(tr, status);
+  tr->tname = tname;
+  set_fork_info(tr, status, expected_signal);
 
   return tr;
 }
 
-static TestResult *receive_result_info_nofork (char *tcname)
+static TestResult *receive_result_info_nofork (const char *tcname,
+                                               const char *tname)
 {
   TestResult *tr;
 
@@ -268,29 +321,49 @@ static TestResult *receive_result_info_nofork (char *tcname)
   if (tr == NULL)
     eprintf("Failed to receive test result", __FILE__, __LINE__);
   tr->tcname = tcname;
+  tr->tname = tname;
   set_nofork_info(tr);
 
   return tr;
 }
 
-static void set_fork_info (TestResult *tr, int status)
+static void set_fork_info (TestResult *tr, int status, int signal_expected)
 {
   int was_sig = WIFSIGNALED(status);
   int was_exit = WIFEXITED(status);
   int exit_status = WEXITSTATUS(status);
+  int signal_received = WTERMSIG(status);
 
   if (was_sig) {
-    tr->rtype = CK_ERROR;
-    tr->msg = signal_msg(WTERMSIG(status));
-  } else if (was_exit && exit_status == 0) {
-    tr->rtype = CK_PASS;
-    tr->msg = pass_msg();
-  } else if (was_exit && exit_status != 0) {
-    if (tr->msg == NULL) { /* early exit */
+    if (signal_expected == signal_received) {
+      tr->rtype = CK_PASS;
+      tr->msg = pass_msg();
+    } else if (signal_expected != 0) { /* signal received, but no the expected one */
       tr->rtype = CK_ERROR;
+      tr->msg = signal_error_msg(signal_received, signal_expected);
+    } else {                           /* signal received and none expected */
+      tr->rtype = CK_ERROR;
+      tr->msg = signal_msg(signal_received);
+    }
+  } else if (signal_expected == 0) {
+    if (was_exit && exit_status == 0) {
+      tr->rtype = CK_PASS;
+      tr->msg = pass_msg();
+    } else if (was_exit && exit_status != 0) {
+      if (tr->msg == NULL) { /* early exit */
+        tr->rtype = CK_ERROR;
+        tr->msg = exit_msg(exit_status);
+      } else {
+        tr->rtype = CK_FAILURE;
+      }
+    }
+  } else { /* a signal was expected and none raised */
+    if (was_exit) {
       tr->msg = exit_msg(exit_status);
-    } else {
-      tr->rtype = CK_FAILURE;
+      if (exit_status == 0)
+	tr->rtype = CK_FAILURE; /* normal exit status */
+      else
+	tr->rtype = CK_FAILURE; /* early exit */
     }
   }
 }
@@ -305,45 +378,72 @@ static void set_nofork_info (TestResult *tr)
   }
 }
 
-static TestResult *tcase_run_tfun_nofork (TCase *tc, TF *tfun)
+static TestResult *tcase_run_tfun_nofork (SRunner *sr, TCase *tc, TF *tfun)
 {
-  tcase_run_checked_setup(tc);
-  tfun->fn();
-  tcase_run_checked_teardown(tc);
-  return receive_result_info_nofork (tc->name);
+  TestResult *tr;
+  
+  tr = tcase_run_checked_setup(sr, tc);
+  if (tr == NULL) {
+    tfun->fn();
+    tcase_run_checked_teardown(tc);
+    return receive_result_info_nofork(tc->name, tfun->name);
+  }
+  
+  return tr;
 }
 
   
-static TestResult *tcase_run_tfun_fork (TCase *tc, TF *tfun)
+static TestResult *tcase_run_tfun_fork (SRunner *sr, TCase *tc, TF *tfun)
 {
   pid_t pid;
   int status = 0;
 
   pid = fork();
   if (pid == -1)
-     eprintf ("Unable to fork:",__FILE__,__LINE__);
+     eprintf("Unable to fork:",__FILE__,__LINE__);
   if (pid == 0) {
-    tcase_run_checked_setup(tc);
+    tcase_run_checked_setup(sr, tc);
     tfun->fn();
     tcase_run_checked_teardown(tc);
-    _exit(EXIT_SUCCESS);
+    exit(EXIT_SUCCESS);
   }
-  (void) wait(&status);
-  return receive_result_info_fork (tc->name, status);
+
+  current_test = pid;
+  alarm(tc->timeout);
+  do {
+    pid = wait(&status);
+  } while (pid == -1);
+
+  return receive_result_info_fork(tc->name, tfun->name, status, tfun->signal);
+}
+
+static char *signal_error_msg (int signal_received, int signal_expected)
+{
+  char *sig_r_str;
+  char *sig_e_str;
+  char *msg = emalloc (MSG_LEN); /* free'd by caller */
+  sig_r_str = strdup(strsignal(signal_received));
+  sig_e_str = strdup(strsignal(signal_expected));
+  snprintf (msg, MSG_LEN, "Error: Received signal %d (%s), expected %d (%s)",
+            signal_received, sig_r_str, signal_expected, sig_e_str);
+  free(sig_r_str);
+  free(sig_e_str);
+  return msg;
 }
 
 static char *signal_msg (int signal)
 {
-  char *msg = emalloc (CMAXMSG); /* free'd by caller */
-  snprintf(msg, CMAXMSG, "Received signal %d", signal);
+  char *msg = emalloc (MSG_LEN); /* free'd by caller */
+  snprintf (msg, MSG_LEN, "Received signal %d (%s)",
+            signal, strsignal(signal));
   return msg;
 }
 
 static char *exit_msg (int exitval)
 {
-  char *msg = emalloc(CMAXMSG); /* free'd by caller */
-  snprintf(msg, CMAXMSG,
-	   "Early exit with return value %d", exitval);
+  char *msg = emalloc(MSG_LEN); /* free'd by caller */
+  snprintf (msg, MSG_LEN,
+            "Early exit with return value %d", exitval);
   return msg;
 }
 
@@ -357,10 +457,10 @@ static char *pass_msg (void)
 enum fork_status srunner_fork_status (SRunner *sr)
 {
   if (sr->fstat == CK_FORK_UNSPECIFIED) {
-    char *env = getenv("CK_FORK");
+    char *env = getenv ("CK_FORK");
     if (env == NULL)
       return CK_FORK;
-    if (strcmp(env,"no") == 0)
+    if (strcmp (env,"no") == 0)
       return CK_NOFORK;
     else
       return CK_FORK;
@@ -373,11 +473,13 @@ void srunner_set_fork_status (SRunner *sr, enum fork_status fstat)
   sr->fstat = fstat;
 }
 
-static int waserror (int status)
+static int waserror (int status, int signal_expected)
 {
-  int was_sig = WIFSIGNALED(status);
-  int was_exit = WIFEXITED(status);
-  int exit_status = WEXITSTATUS(status);
+  int was_sig = WIFSIGNALED (status);
+  int was_exit = WIFEXITED (status);
+  int exit_status = WEXITSTATUS (status);
+  int signal_received = WTERMSIG(status);
 
-  return (was_sig || (was_exit && exit_status != 0));
+  return ((was_sig && (signal_received != signal_expected)) ||
+          (was_exit && exit_status != 0));
 }
